import os
from typing import Optional, Tuple, List, Union, Callable

import numpy as np
import torch
from torch import nn
import matplotlib.pyplot as plt

from tqdm import trange
import torch.nn.functional as F


# For repeatability
seed = 0
torch.manual_seed(seed)
np.random.seed(seed)
# print(torch.__version__)
device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
print("running on device", device)
data = np.load('tiny_nerf_data.npz')


near, far = 2., 6.

n_training = 100
testimg_idx = 101



# Gather as torch tensors
images = torch.from_numpy(data['images'][:n_training]).to(device)
height, width = images.shape[1:3]
images = torch.norm(images, dim=-1)
images[images!=0] = 1 # convert to 1-0 masks
print(images.shape) #torch.Size([100, 100, 100])

poses = torch.from_numpy(data['poses']).to(device)
focal = torch.from_numpy(data['focal']).to(device)
testimg = torch.from_numpy(data['images'][testimg_idx]).to(device)
testpose = torch.from_numpy(data['poses'][testimg_idx]).to(device)

testimg = torch.norm(testimg, dim=-1)
testimg[testimg!=0] = 1 # convert to 1-0 masks


d_input = 3          
n_freqs = 10 

# Stratified sampling
n_samples = 64*8        
#n_samples = 64*4
perturb = False         # If set, applies noise to sample positions
inverse_depth = False  # If set, samples points linearly in inverse depth

useEncoder = True

# Optimizer
lr = 5e-4  # Learning rate

# Training
n_iters = 10000
batch_size = 2**14          # Number of rays per gradient step (power of 2)
#batch_size = 2**10 
one_image_per_step = True   # One image per gradient step (disables batching)
chunksize = 2**12           # Modify as needed to fit in GPU memory
#chunksize = 2**8

display_rate = 500          # Display test output every X epochs


# We bundle the kwargs for various functions to pass all at once.
kwargs_sample_stratified = {
    'n_samples': n_samples,
    'perturb': perturb,
    'inverse_depth': inverse_depth
}


def get_rays(
  height: int,
  width: int,
  focal_length: float,
  c2w: torch.Tensor
) -> Tuple[torch.Tensor, torch.Tensor]:
  r"""
  Find origin and direction of rays through every pixel and camera origin.
  """

  # Apply pinhole camera model to gather directions at each pixel
  i, j = torch.meshgrid(
      torch.arange(width, dtype=torch.float32).to(c2w),
      torch.arange(height, dtype=torch.float32).to(c2w),
      indexing='ij')
  i, j = i.transpose(-1, -2), j.transpose(-1, -2)
  directions = torch.stack([(i - width * .5) / focal_length,
                            -(j - height * .5) / focal_length,
                            -torch.ones_like(i)
                           ], dim=-1)

  # Apply camera pose to directions
  rays_d = torch.sum(directions[..., None, :] * c2w[:3, :3], dim=-1)

  # Origin is same for all directions (the optical center)
  rays_o = c2w[:3, -1].expand(rays_d.shape)
  return rays_o, rays_d





def sample_stratified(
  rays_o: torch.Tensor,
  rays_d: torch.Tensor,
  near: float,
  far: float,
  n_samples: int,
  perturb: Optional[bool] = True,
  inverse_depth: bool = False
) -> Tuple[torch.Tensor, torch.Tensor]:
  r"""
  Sample along ray from regularly-spaced bins.
  """

  # Grab samples for space integration along ray
  t_vals = torch.linspace(0., 1., n_samples, device=rays_o.device)
  if not inverse_depth:
    # Sample linearly between `near` and `far`
    z_vals = near * (1.-t_vals) + far * (t_vals)
  else:
    # Sample linearly in inverse depth (disparity)
    z_vals = 1./(1./near * (1.-t_vals) + 1./far * (t_vals))

  # Draw uniform samples from bins along ray
  if perturb:
    mids = .5 * (z_vals[1:] + z_vals[:-1])
    upper = torch.concat([mids, z_vals[-1:]], dim=-1)
    lower = torch.concat([z_vals[:1], mids], dim=-1)
    t_rand = torch.rand([n_samples], device=z_vals.device)
    z_vals = lower + (upper - lower) * t_rand
  z_vals = z_vals.expand(list(rays_o.shape[:-1]) + [n_samples])

  # Apply scale from `rays_d` and offset from `rays_o` to samples
  # pts: (width, height, n_samples, 3)
  pts = rays_o[..., None, :] + rays_d[..., None, :] * z_vals[..., :, None]
  return pts, z_vals







class PositionalEncoder(nn.Module):
    """
    Sine-cosine positional encoder for input points.
    """
    def __init__(self, d_input: int, n_freqs: int, log_space: bool = False):

        super().__init__()
        self.d_input = d_input
        self.n_freqs = n_freqs
        self.log_space = log_space
        self.d_output = d_input * (1 + 2 * self.n_freqs)
        self.embed_fns = [lambda x: x]

        # Define frequencies in either linear or log scale
        if self.log_space:
            freq_bands = 2.**torch.linspace(0., self.n_freqs - 1, self.n_freqs)
        else:
            freq_bands = torch.linspace(2.**0., 2.**(self.n_freqs - 1), self.n_freqs)

        # Alternate sin and cos
        for freq in freq_bands:
            self.embed_fns.append(lambda x, freq=freq: torch.sin(x * freq))
            self.embed_fns.append(lambda x, freq=freq: torch.cos(x * freq))

    def forward(self, x) -> torch.Tensor:
        """
        Apply positional encoding to input.
        """
        return torch.concat([fn(x) for fn in self.embed_fns], dim=-1)


def get_chunks(inputs: torch.Tensor, chunksize: int = 2**15) -> List[torch.Tensor]:
    """
    Divide an input into chunks.
    """
    return [inputs[i:i + chunksize] for i in range(0, inputs.shape[0], chunksize)]

def prepare_chunks(points: torch.Tensor, encoding_function: Callable[[torch.Tensor], torch.Tensor], chunksize: int = 2**15) -> List[torch.Tensor]:
    """
    Encode and chunkify points to prepare for NeRF model.
    """
    # [10000, 512, 3] points
    d = points.shape[-1]
    points = points.reshape((-1, d))
    # [10000*512, 3] 
    if encoding_function is not None:
        points = encoding_function(points)
    points = get_chunks(points, chunksize=chunksize)
    return points



def nerf_forward(
  rays_o: torch.Tensor,
  rays_d: torch.Tensor,
  near: float,
  far: float,
  target_img: torch.Tensor,
  nerf_model: nn.Module,
  encoding_fn: Callable[[torch.Tensor], torch.Tensor],
  W: torch.Tensor = None,
  kwargs_sample_stratified: dict = None,
  chunksize: int = 2**15
) -> Tuple[torch.Tensor, torch.Tensor]:
  r"""
  Compute forward pass through model(s).
  """

  # Set no kwargs if none are given.
  if kwargs_sample_stratified is None:
    kwargs_sample_stratified = {}


  # Sample query points along each ray.
  query_points, z_vals = sample_stratified(
      rays_o, rays_d, near, far, **kwargs_sample_stratified)


  # Prepare batches.
  batches = prepare_chunks(query_points, encoding_function=encoding_fn,chunksize=chunksize)

  predictions = []

  for batch in batches:

    pred = nerf_model(batch)
    #print("pred size for each chunk: ", pred.shape) #torch.Size([4096, 1])
    predictions.append(pred)
  raw = torch.cat(predictions, dim=0).unsqueeze(-1)
  # print(raw.shape) #torch.Size([5120000, 1])
  raw = raw.reshape(list(query_points.shape[:2]) + [raw.shape[-1]]).squeeze(-1)
  #print(raw.shape) # torch.Size([10000, 512, 1]

  return (raw) 






class Nerf(nn.Module):
    def __init__(self, in_features, hid_features, out_features):
        super(Nerf, self).__init__()

        self.fc1 = nn.Linear(in_features, hid_features)
        self.fc2 = nn.Linear(hid_features, hid_features)
        self.fc3 = nn.Linear(hid_features, out_features)

    def forward(self, x):


        x = F.relu(self.fc1(x))
        z = F.relu(self.fc2(x))
        x = self.fc3(z)

        return x





def init_models():
  r"""
  Initialize models, encoders, and optimizer for NeRF training.
  """
  # Models
  # in_features, hid_features, out_features, nonlin
  if useEncoder == True:

    encoder = PositionalEncoder(d_input, n_freqs, log_space=True)
    encode = lambda x: encoder(x)
    in_features = encoder.d_output

  else:
    encode = None
    in_features = 3
  

  model = Nerf(in_features=in_features, hid_features=256, out_features=1)
  
  model.to(device)
  model_params = list(model.parameters())

  optimizer = torch.optim.AdamW(model_params, lr=lr, betas=(0.8, 0.999)) 

  return model, optimizer, encode





def render_image(raw):
  # raw.shape = #([10000, 512]) (H*W, num_samples_along_each_ray)
  image = torch.mean(raw, dim=-1)
  return image

def train(images, testimg):
  r"""
  Launch training session for NeRF.
  """

  train_psnrs = []
  val_psnrs = []
  iternums = []
  for i in trange(n_iters):
    model.train()

    if one_image_per_step:
      # Randomly pick an image as the target.
      target_img_idx = np.random.randint(images.shape[0])
      target_img = images[target_img_idx].to(device)

      height, width = target_img.shape[:2]
      target_pose = poses[target_img_idx].to(device)
      rays_o, rays_d = get_rays(height, width, focal, target_pose)

      W = None
      rays_o = rays_o.reshape([-1, 3])
      rays_d = rays_d.reshape([-1, 3]) # H*W, 3

    # print(target_img.shape) #torch.Size([100, 100])
    target_img = target_img.reshape([-1]) # H*W
    # print(target_img.shape) #torch.Size([10000])

    (raw) = nerf_forward(rays_o, rays_d,
                           near, far, target_img, model, encode, W=W,
                           kwargs_sample_stratified=kwargs_sample_stratified,
                           chunksize=chunksize)

    # Check for any numerical issues.
    
    if torch.isnan(raw).any():
        print(f"! [Numerical Alert] predictions contain NaN.")
    if torch.isinf(raw).any():
        print(f"! [Numerical Alert] predictions contain Inf.")

    pred_img = render_image(raw)
    imgLoss = torch.nn.functional.mse_loss(pred_img, target_img)

    imgLoss.backward() #trying the other loss now
    optimizer.step()
    optimizer.zero_grad()
    
    psnr = -10. * torch.log10(imgLoss)
    train_psnrs.append(psnr.item())

    # Evaluate testimg at given display rate.
    if ((i) % display_rate == 0) or (i == n_iters-1):
      model.eval()

      rays_o, rays_d = get_rays(height, width, focal, testpose)

      W = None
      
      rays_o = rays_o.reshape([-1, 3])
      rays_d = rays_d.reshape([-1, 3])
      testimg = testimg.reshape([-1])
      (raw) = nerf_forward(rays_o, rays_d,
                           near, far, testimg, model, encode, W=W,
                           kwargs_sample_stratified=kwargs_sample_stratified,
                           chunksize=chunksize)
      

      pred_img = render_image(raw)
      imgLoss = torch.nn.functional.mse_loss(pred_img, testimg)
      print("Loss:", imgLoss.item())
      val_psnr = -10. * torch.log10(imgLoss)
      val_psnrs.append(val_psnr.item())
      iternums.append(i)

      # Plot example outputs
      fig, ax = plt.subplots(1, 3, figsize=(12,4), gridspec_kw={'width_ratios': [1, 1, 1]})
      ax[0].imshow(pred_img.reshape([height, width]).detach().cpu().numpy())
      ax[0].set_title(f'Iteration: {i}')
      ax[1].imshow(testimg.reshape([height, width]).detach().cpu().numpy())
      ax[1].set_title(f'Target')
      ax[2].plot(range(0, i + 1), train_psnrs, 'r')
      ax[2].plot(iternums, val_psnrs, 'b')
      ax[2].set_title('PSNR (train=red, val=blue')

      savedir = "/home/irmak/convex_nerf/new_updt1/"
      os.makedirs(savedir, exist_ok=True)
      plt.savefig(savedir + "val_outs_iter_{}.png".format(i))
      plt.close()

      print("PSNR")
      print(val_psnrs[-1], flush=True)
      np.save(savedir + "val_psnrs.npy", val_psnrs)
      np.save(savedir + "train_psnrs.npy", train_psnrs)
      torch.save(model.state_dict(), savedir + 'model.pt')
      




  return True, train_psnrs, val_psnrs

if __name__=="__main__":


    model, optimizer, encode = init_models()

    success, train_psnrs, val_psnrs = train(images, testimg)


    print('')
    print(f'Done!')

    print("VAL PSNR")
    print(val_psnrs[-1])



