import os
from typing import Optional, Tuple, List, Union, Callable

import numpy as np
import torch
from torch import nn
import matplotlib.pyplot as plt

from tqdm import trange
import torch.nn.functional as F







def get_rays(
  height: int,
  width: int,
  focal_length: float,
  c2w: torch.Tensor
) -> Tuple[torch.Tensor, torch.Tensor]:
  r"""
  Find origin and direction of rays through every pixel and camera origin.
  """

  # Apply pinhole camera model to gather directions at each pixel
  i, j = torch.meshgrid(
      torch.arange(width, dtype=torch.float32).to(c2w),
      torch.arange(height, dtype=torch.float32).to(c2w),
      indexing='ij')
  i, j = i.transpose(-1, -2), j.transpose(-1, -2)
  directions = torch.stack([(i - width * .5) / focal_length,
                            -(j - height * .5) / focal_length,
                            -torch.ones_like(i)
                           ], dim=-1)

  # Apply camera pose to directions
  rays_d = torch.sum(directions[..., None, :] * c2w[:3, :3], dim=-1)

  # Origin is same for all directions (the optical center)
  rays_o = c2w[:3, -1].expand(rays_d.shape)
  return rays_o, rays_d





def sample_stratified(
  rays_o: torch.Tensor,
  rays_d: torch.Tensor,
  near: float,
  far: float,
  n_samples: int,
  perturb: Optional[bool] = True,
  inverse_depth: bool = False
) -> Tuple[torch.Tensor, torch.Tensor]:
  r"""
  Sample along ray from regularly-spaced bins.
  """

  # Grab samples for space integration along ray
  t_vals = torch.linspace(0., 1., n_samples, device=rays_o.device)
  if not inverse_depth:
    # Sample linearly between `near` and `far`
    z_vals = near * (1.-t_vals) + far * (t_vals)
  else:
    # Sample linearly in inverse depth (disparity)
    z_vals = 1./(1./near * (1.-t_vals) + 1./far * (t_vals))

  # Draw uniform samples from bins along ray
  if perturb:
    mids = .5 * (z_vals[1:] + z_vals[:-1])
    upper = torch.concat([mids, z_vals[-1:]], dim=-1)
    lower = torch.concat([z_vals[:1], mids], dim=-1)
    t_rand = torch.rand([n_samples], device=z_vals.device)
    z_vals = lower + (upper - lower) * t_rand
  z_vals = z_vals.expand(list(rays_o.shape[:-1]) + [n_samples])

  # Apply scale from `rays_d` and offset from `rays_o` to samples
  # pts: (width, height, n_samples, 3)
  pts = rays_o[..., None, :] + rays_d[..., None, :] * z_vals[..., :, None]
  return pts, z_vals







class PositionalEncoder(nn.Module):
    """
    Sine-cosine positional encoder for input points.
    """
    def __init__(self, d_input: int, n_freqs: int, log_space: bool = False):

        super().__init__()
        self.d_input = d_input
        self.n_freqs = n_freqs
        self.log_space = log_space
        self.d_output = d_input * (1 + 2 * self.n_freqs)
        self.embed_fns = [lambda x: x]

        # Define frequencies in either linear or log scale
        if self.log_space:
            freq_bands = 2.**torch.linspace(0., self.n_freqs - 1, self.n_freqs)
        else:
            freq_bands = torch.linspace(2.**0., 2.**(self.n_freqs - 1), self.n_freqs)

        # Alternate sin and cos
        for freq in freq_bands:
            self.embed_fns.append(lambda x, freq=freq: torch.sin(x * freq))
            self.embed_fns.append(lambda x, freq=freq: torch.cos(x * freq))

    def forward(self, x) -> torch.Tensor:
        """
        Apply positional encoding to input.
        """
        return torch.concat([fn(x) for fn in self.embed_fns], dim=-1)



def map_pose_to_proj(ray_pts, ref_lbls, D):

  N = ref_lbls.shape[0]

  xcoord = ray_pts[:, 0]
  ycoord = ray_pts[:, 1]
  zcoord = ray_pts[:, 2]

  k = torch.round((xcoord + D/2) * N/D).to(torch.int)
  l = torch.round((ycoord + D/2) * N/D).to(torch.int)
  m = torch.round((zcoord + D/2) * N/D).to(torch.int)

  cond = lambda x: torch.logical_and(x >= 0, x < N)

  cond_all = torch.logical_and(cond(k), torch.logical_and(cond(l), cond(m)))
  valid_all = torch.where(cond_all)


  k_valid = k[valid_all]
  l_valid = l[valid_all]
  m_valid = m[valid_all]

  proj = ref_lbls[k_valid,l_valid,m_valid]
  return proj


def get_chunks(inputs: torch.Tensor, chunksize: int = 2**15) -> List[torch.Tensor]:
    """
    Divide an input into chunks.
    """
    return [inputs[i:i + chunksize] for i in range(0, inputs.shape[0], chunksize)]

def prepare_chunks(points: torch.Tensor, labels: torch.Tensor, encoding_function: Callable[[torch.Tensor], torch.Tensor], chunksize: int = 2**15) -> List[torch.Tensor]:
    """
    Encode and chunkify points to prepare for NeRF model.
    """
    # [10000, 512, 3] points
    d = points.shape[-1]
    points = points.reshape((-1, d))
    # [10000*512, 3] 
    if encoding_function is not None:
        points = encoding_function(points)
    points = get_chunks(points, chunksize=chunksize)
    labels = get_chunks(labels, chunksize=chunksize)
    return points, labels


def nerf_forward(ref_volume, carved_labels, nerf_model, encoding_fn, chunksize):
  N = carved_labels.shape[0]
  batches, batched_labels = prepare_chunks(
    ref_volume.reshape([-1, 3]), carved_labels.flatten(), encoding_function=encoding_fn,chunksize=chunksize)
  loss = 0.
  preds = []
  for i, batch in enumerate(batches):
    # print(batch.shape)
    # torch.Size([4096, 63])
    lbl = batched_labels[i]
    pred = nerf_model(batch)
    preds.append(pred)
    loss += torch.nn.functional.mse_loss(pred.squeeze(), lbl)

  sc_loss = loss / len(batches)
  preds = torch.cat(preds, dim=0).unsqueeze(-1)
  # print(preds.shape)
  preds = preds.reshape([N,N,N])
  return (sc_loss, preds)


def render_nerf(nerf_labels):

    rays_o, rays_d = get_rays(height, width, focal, testpose) # H, W, 3

    rays_o = rays_o.reshape([-1, 3])
    rays_d = rays_d.reshape([-1, 3]) # H*W, 3

    query_points, z_vals = sample_stratified(rays_o, rays_d, near, far, **kwargs_sample_stratified)
      # print(query_points.shape) # torch.Size([10000, 512, 3])
    recon = torch.zeros(height*width).to(device)
    for pix in range(query_points.shape[0]):
      ray = query_points[pix, :, :]
      proj = map_pose_to_proj(ray, nerf_labels, D)
      recon[pix] = torch.sum(proj)

    return recon.reshape([height, width])


class Nerf(nn.Module):
    def __init__(self, in_features, hid_features, out_features):
        super(Nerf, self).__init__()

        self.fc1 = nn.Linear(in_features, hid_features)
        self.fc2 = nn.Linear(hid_features, hid_features)
        self.fc3 = nn.Linear(hid_features, out_features)


    def forward(self, x):

        x = F.relu(self.fc1(x))
        z = F.relu(self.fc2(x))
        x = self.fc3(z)
        return x





def init_models():
  r"""
  Initialize models, encoders, and optimizer for NeRF training.
  """

  if useEncoder == True:
    encoder = PositionalEncoder(d_input, n_freqs, log_space=True)
    encode = lambda x: encoder(x)
    in_features = encoder.d_output
  else:
    encode = None
    in_features = 3


  model = Nerf(in_features=in_features, hid_features=256, out_features=1)
  
  model.to(device)
  model_params = list(model.parameters())

  # Optimizer
  # optimizer = torch.optim.Adam(model_params, lr=lr)
  optimizer = torch.optim.AdamW(model_params, lr=lr)

  return model, optimizer, encode

def get_psnr(x, gt):
  x = x/torch.max(x)
  gt = gt/torch.max(gt)
  out = 10 * torch.log10(1 / ((x - gt)**2).mean())
  return out

def train(images, testimg):

  train_losses = []
  val_psnrs = []
  iternums = []
  for i in trange(n_iters):
    model.train()


    (sc_loss, preds) = nerf_forward(ref_coords, carved_labels, model, encode, chunksize)


    sc_loss.backward() 
    optimizer.step()
    optimizer.zero_grad()
    
    train_losses.append(sc_loss.item())
    print(sc_loss)

    # Evaluate testimg at given display rate.
    if ((i) % display_rate == 0) or (i == n_iters-1):
      model.eval()

      rays_o, rays_d = get_rays(height, width, focal, testpose)

      (sc_loss, preds) = nerf_forward(ref_coords, carved_labels, model, encode, chunksize)
      recon = render_nerf(preds)

      
      imgLoss = torch.nn.functional.mse_loss(recon, testimg)
      print("3D SC Loss:", sc_loss.item())
      print("2D Image Loss:", imgLoss.item())
      val_psnr = get_psnr(recon.flatten(), testimg.flatten())
      val_psnrs.append(val_psnr.item())
      iternums.append(i)

      # Plot example outputs
      fig, ax = plt.subplots(1, 3, figsize=(12,4), gridspec_kw={'width_ratios': [1, 1, 1]})
      ax[0].imshow(recon.detach().cpu().numpy())
      ax[0].set_title(f'Iteration: {i}')
      ax[1].imshow(testimg.reshape([height, width]).detach().cpu().numpy())
      ax[1].set_title(f'SC GT')
      ax[2].plot(iternums, val_psnrs)
      ax[2].set_title('PSNR (val)')
      #/home/irmak/convex_nerf/nerf_full.py
      savedir = "/home/irmak/convex_nerf/sc_nerf_4/"
      os.makedirs(savedir, exist_ok=True)
      plt.savefig(savedir + "val_outs_iter_{}.png".format(i))
      plt.close()

      print("PSNR")
      print(val_psnrs[-1], flush=True)
      np.save(savedir + "val_psnrs.npy", val_psnrs)
      # np.save(savedir + "train_psnrs.npy", train_psnrs)
      torch.save(model.state_dict(), savedir + 'model.pt')
      


  return train_losses, val_psnrs

if __name__=="__main__":

  # For repeatability
  seed = 0
  torch.manual_seed(seed)
  np.random.seed(seed)
  # print(torch.__version__)
  device = torch.device('cuda:1' if torch.cuda.is_available() else 'cpu')
  print("running on device", device)
  data = np.load('tiny_nerf_data.npz')


  near, far = 2., 6.

  n_training = 100
  testimg_idx = 101



  # Gather as torch tensors
  images = torch.from_numpy(data['images'][:n_training]).to(device)
  height, width = images.shape[1:3]
  images = torch.norm(images, dim=-1)
  images[images!=0] = 1 # convert to 1-0 masks
  print(images.shape) #torch.Size([100, 100, 100])

  poses = torch.from_numpy(data['poses']).to(device)
  focal = torch.from_numpy(data['focal']).to(device)
  testimg = torch.from_numpy(data['images'][testimg_idx]).to(device)
  testpose = torch.from_numpy(data['poses'][testimg_idx]).to(device)

  testimg = torch.norm(testimg, dim=-1)
  testimg[testimg!=0] = 1 # convert to 1-0 masks
  gt = np.load("ground_truth_SC.npy")
  gt = torch.from_numpy(gt).to(device)




  d_input = 3           # Number of input dimensions
  n_freqs = 10 
  mapping_fn = "positional_enc"
  chunksize = 2**12 
  batch_size = 2**14
  # Stratified sampling
  n_samples = 64*8        # Number of spatial samples per ray
  #n_samples = 64*4
  perturb = False         # If set, applies noise to sample positions
  inverse_depth = False  # If set, samples points linearly in inverse depth

  useEncoder = True

  # Optimizer
  lr = 5e-4  # Learning rate

  # Training
  n_iters = 3000 
  one_image_per_step = True   # One image per gradient step (disables batching)

  display_rate = 50          # Display test output every X epochs


  kwargs_sample_stratified = {
      'n_samples': n_samples,
      'perturb': perturb,
      'inverse_depth': inverse_depth
  }

  carved_labels = torch.load("sc_vol_N_256.pt").to(device)
  carved_labels = carved_labels[25:-25, 25:-25, 25:-25]
  ts = poses[:,:3,-1] # torch.Size([106, 3])
  D = torch.mean(torch.norm(ts,dim=-1))
  D = D*0.8

  # ref coords
  N = carved_labels.shape[0]
  xr = torch.linspace(-D/2, D/2, N)
  yr = torch.linspace(D/2, -D/2, N)
  zr = torch.linspace(-D/2, D/2, N)
  xx, yy, zz = torch.meshgrid(xr, yr, zr, indexing='ij')

  ref_coords = torch.stack([xx,yy,zz]).permute([1,2,3,0]).to(device) # torch.Size([512, 512, 512, 3])

  model, optimizer, encode = init_models()

  train_psnrs, val_psnrs = train(images, gt)


  print('')
  print(f'Done!')

  print("VAL PSNR")
  print(val_psnrs[-1])


