from typing import Optional, Tuple, List, Union, Callable

import numpy as np
import torch
import time
import matplotlib.pyplot as plt

# For repeatability
seed = 0
torch.manual_seed(seed)
np.random.seed(seed)

device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
print("running on device", device)



def get_rays(
  height: int,
  width: int,
  focal_length: float,
  c2w: torch.Tensor
) -> Tuple[torch.Tensor, torch.Tensor]:
  r"""
  Find origin and direction of rays through every pixel and camera origin.
  """

  # Apply pinhole camera model to gather directions at each pixel
  i, j = torch.meshgrid(
      torch.arange(width, dtype=torch.float32).to(c2w),
      torch.arange(height, dtype=torch.float32).to(c2w),
      indexing='ij')
  i, j = i.transpose(-1, -2), j.transpose(-1, -2)
  directions = torch.stack([(i - width * .5) / focal_length,
                            -(j - height * .5) / focal_length,
                            -torch.ones_like(i)
                           ], dim=-1)

  # Apply camera pose to directions
  rays_d = torch.sum(directions[..., None, :] * c2w[:3, :3], dim=-1)

  # Origin is same for all directions (the optical center)
  rays_o = c2w[:3, -1].expand(rays_d.shape)
  return rays_o, rays_d





def sample_stratified(
  rays_o: torch.Tensor,
  rays_d: torch.Tensor,
  near: float,
  far: float,
  n_samples: int,
  perturb: Optional[bool] = True,
  inverse_depth: bool = False
) -> Tuple[torch.Tensor, torch.Tensor]:
  r"""
  Sample along ray from regularly-spaced bins.
  """

  # Grab samples for space integration along ray
  t_vals = torch.linspace(0., 1., n_samples, device=rays_o.device)
  if not inverse_depth:
    # Sample linearly between `near` and `far`
    z_vals = near * (1.-t_vals) + far * (t_vals)
  else:
    # Sample linearly in inverse depth (disparity)
    z_vals = 1./(1./near * (1.-t_vals) + 1./far * (t_vals))

  # Draw uniform samples from bins along ray
  if perturb:
    mids = .5 * (z_vals[1:] + z_vals[:-1])
    upper = torch.concat([mids, z_vals[-1:]], dim=-1)
    lower = torch.concat([z_vals[:1], mids], dim=-1)
    t_rand = torch.rand([n_samples], device=z_vals.device)
    z_vals = lower + (upper - lower) * t_rand
  z_vals = z_vals.expand(list(rays_o.shape[:-1]) + [n_samples])

  # Apply scale from `rays_d` and offset from `rays_o` to samples
  # pts: (width, height, n_samples, 3)
  pts = rays_o[..., None, :] + rays_d[..., None, :] * z_vals[..., :, None]
  return pts, z_vals



def map_rays_to_ref(ray_pts, ref_lbls, D):
  # input: q,3 samples along one ray (the pixel value of it is 0)
  # output: sets relevant ref points to zero
  N = ref_lbls.shape[0]

  xcoord = ray_pts[:, 0]
  ycoord = ray_pts[:, 1]
  zcoord = ray_pts[:, 2]

  k = torch.round((xcoord + D/2) * N/D).to(torch.int)
  l = torch.round((ycoord + D/2) * N/D).to(torch.int)
  m = torch.round((zcoord + D/2) * N/D).to(torch.int)

  cond = lambda x: torch.logical_and(x >= 0, x < N)

  cond_all = torch.logical_and(cond(k), torch.logical_and(cond(l), cond(m)))
  valid_all = torch.where(cond_all)
  # print(len(valid_all))
  # print(len(valid_all[0]))

  k_valid = k[valid_all]
  l_valid = l[valid_all]
  m_valid = m[valid_all]

  ref_lbls[k_valid,l_valid,m_valid] = 0
  return ref_lbls


def map_pose_to_proj(ray_pts, ref_lbls, D):
  # input: q,3 samples along one ray (the pixel value of it is 0)
  # output: sets relevant ref points to zero
  N = ref_lbls.shape[0]

  xcoord = ray_pts[:, 0]
  ycoord = ray_pts[:, 1]
  zcoord = ray_pts[:, 2]

  k = torch.round((xcoord + D/2) * N/D).to(torch.int)
  l = torch.round((ycoord + D/2) * N/D).to(torch.int)
  m = torch.round((zcoord + D/2) * N/D).to(torch.int)

  cond = lambda x: torch.logical_and(x >= 0, x < N)

  cond_all = torch.logical_and(cond(k), torch.logical_and(cond(l), cond(m)))
  valid_all = torch.where(cond_all)


  k_valid = k[valid_all]
  l_valid = l[valid_all]
  m_valid = m[valid_all]

  proj = ref_lbls[k_valid,l_valid,m_valid]
  return proj


def carving(images):

  current_lbls = ref_labels
  for i in range(images.shape[0]):

    target_img = images[i].to(device)

    height, width = target_img.shape[:2]
    target_pose = poses[i].to(device)
    rays_o, rays_d = get_rays(height, width, focal, target_pose) # H, W, 3

    rays_o = rays_o.reshape([-1, 3])
    rays_d = rays_d.reshape([-1, 3]) # H*W, 3


    target_img = target_img.reshape([-1]) # H*W


    query_points, z_vals = sample_stratified(
      rays_o, rays_d, near, far, **kwargs_sample_stratified) # torch.Size([10000, 512, 3])
  
    pix_0 = torch.nonzero(target_img==0)
    ray_0 = query_points[pix_0,:,:] # 0 ray coordinates


    new_labels = map_rays_to_ref(ray_0.reshape([-1, 3]), current_lbls, D)
    current_lbls = new_labels
  
  return current_lbls
    

def get_psnr(x, gt):
  x = x/np.max(x.flatten(),axis=0)
  gt = gt/np.max(gt.flatten(),axis=0)
  out = 10 * np.log10(1 / ((x - gt)**2).mean().item())
  return out


if __name__=="__main__":
  data = np.load('tiny_nerf_data.npz')

  near, far = 2., 6.
  useTrain = True

  if useTrain:
    n_training = 100
  else:
    n_training = 106

  testimg_idx = 101


  images = torch.from_numpy(data['images'][:n_training]).to(device)
  height, width = images.shape[1:3]
  images = torch.norm(images, dim=-1)
  images[images!=0] = 1 # convert to 1-0 masks


  poses = torch.from_numpy(data['poses']).to(device)
  focal = torch.from_numpy(data['focal']).to(device)
  testimg = torch.from_numpy(data['images'][testimg_idx]).to(device)
  testpose = torch.from_numpy(data['poses'][testimg_idx]).to(device)

  testimg = torch.norm(testimg, dim=-1)
  testimg[testimg!=0] = 1 # convert to 1-0 masks


  ts = poses[:,:3,-1] # torch.Size([106, 3])
  D = torch.mean(torch.norm(ts,dim=-1))
  # print(D) # 4.0311

  # Stratified sampling
  n_samples = 64*8        # Number of spatial samples per ray
  perturb = False         # If set, applies noise to sample positions
  inverse_depth = False  # If set, samples points linearly in inverse depth


  kwargs_sample_stratified = {
      'n_samples': n_samples,
      'perturb': perturb,
      'inverse_depth': inverse_depth
  }

  # ref coords
  # xr = torch.linspace(-D/2, D/2, n_samples)
  # yr = torch.linspace(D/2, -D/2, n_samples)
  # zr = torch.linspace(-D/2, D/2, n_samples)
  # xx, yy, zz = torch.meshgrid(xr, yr, zr, indexing='ij')

  # ref_coords = torch.stack([xx,yy,zz]).permute([1,2,3,0]).to(device) # torch.Size([512, 512, 512, 3])
 
  # ref labels
  if useTrain:
    #Ns = [256] #[64, 128, 256, 512]
    N = 256
    gt = np.load("ground_truth_SC.npy")
    Ds = [D] #[D*0.5, D*0.75, D*0.85, D*0.95, D]
    psnrs = []
    t_start = time.time()
    for D in Ds:
      ref_labels = torch.ones(size=(N,N,N)).to(device)
      carved_labels = carving(images)
      # NOTE: uncomment here to save the desired parameter SC volume!
      # torch.save(carved_labels, "sc_vol_N_{}.pt".format(N))
      # torch.save(carved_labels, "sc_vol_D_{}.pt".format(D))
      # print(carved_labels.shape)

      height, width = testimg.shape[:2]

      rays_o, rays_d = get_rays(height, width, focal, testpose) # H, W, 3

      rays_o = rays_o.reshape([-1, 3])
      rays_d = rays_d.reshape([-1, 3]) # H*W, 3

      query_points, z_vals = sample_stratified(
        rays_o, rays_d, near, far, **kwargs_sample_stratified)
      # print(query_points.shape) # torch.Size([10000, 512, 3])
      recon = torch.zeros(height*width)
      for pix in range(query_points.shape[0]):
        ray = query_points[pix, :, :]
        proj = map_pose_to_proj(ray, carved_labels, D)
        recon[pix] = torch.sum(proj)
      recon = recon.reshape([height, width])
      recon = recon.detach().cpu().numpy()
      psnrs.append(get_psnr(recon, gt))
      # plt.figure()
      # plt.imshow(recon)
      # plt.savefig("testpose_D_{}.png".format(D))
      #np.save("SC_N_{}.npy".format(N), recon)
    t_end = time.time()
    print(t_end-t_start)
    print(psnrs)
    # plt.figure()
    # # plt.plot(Ns, psnrs)
    # D = D.detach().cpu().numpy()
    # Ds = [D*0.5, D*0.75, D*0.85, D*0.95, D]
    # plt.plot(Ds, psnrs)
    # # plt.xlabel("N")
    # plt.xlabel("D")
    # plt.ylabel("PSNR")
    # plt.savefig("psnr_vs_D.png")
  else:
    N = 256
    ref_labels = torch.ones(size=(N,N,N)).to(device)
    carved_labels = carving(images)
    torch.save(carved_labels, "gt_sc_vol.pt")
    # print(carved_labels.shape)

    height, width = testimg.shape[:2]

    rays_o, rays_d = get_rays(height, width, focal, testpose) # H, W, 3

    rays_o = rays_o.reshape([-1, 3])
    rays_d = rays_d.reshape([-1, 3]) # H*W, 3

    query_points, z_vals = sample_stratified(
      rays_o, rays_d, near, far, **kwargs_sample_stratified)
      # print(query_points.shape) # torch.Size([10000, 512, 3])
    recon = torch.zeros(height*width)
    for pix in range(query_points.shape[0]):
      ray = query_points[pix, :, :]
      proj = map_pose_to_proj(ray, carved_labels, D)
      recon[pix] = torch.sum(proj)
    recon = recon.reshape([height, width])
    recon = recon.detach().cpu().numpy()
    np.save("ground_truth_SC.npy", recon)


  # NOTE: can use these to plot SC projections along x,y,z axes.
    
  # fig, ax = plt.subplots(1, 3, figsize=(12,4), gridspec_kw={'width_ratios': [1, 1, 1]})
  # ax[0].imshow(carved_labels[:,:,25].detach().cpu().numpy())
  # ax[0].set_title('y=50')
  # ax[1].imshow(carved_labels[:,:,50].detach().cpu().numpy())
  # ax[1].set_title('y=100')
  # ax[2].imshow(carved_labels[:,:,75].detach().cpu().numpy())
  # ax[2].set_title('y=150')
  # plt.savefig("carve_test11.png")

  # fig, ax = plt.subplots(1, 3, figsize=(12,4), gridspec_kw={'width_ratios': [1, 1, 1]})
  # ax[0].imshow(carved_labels[:,25,:].detach().cpu().numpy())
  # ax[0].set_title('y=200')
  # ax[1].imshow(carved_labels[:,50,:].detach().cpu().numpy())
  # ax[1].set_title('y=250')
  # ax[2].imshow(carved_labels[:,75,:].detach().cpu().numpy())
  # ax[2].set_title('y=300')
  # plt.savefig("carve_test12.png")

  # fig, ax = plt.subplots(1, 3, figsize=(12,4), gridspec_kw={'width_ratios': [1, 1, 1]})
  # ax[0].imshow(carved_labels[25,:,:].detach().cpu().numpy())
  # ax[0].set_title('y=350')
  # ax[1].imshow(carved_labels[50,:,:].detach().cpu().numpy())
  # ax[1].set_title('y=400')
  # ax[2].imshow(carved_labels[75,:,:].detach().cpu().numpy())
  # ax[2].set_title('y=450')
  # plt.savefig("carve_test13.png")

# psnr of SC:
# 38.96690926806394

  

















